@@ -166,7 +166,12 @@ perlinstmod () {
       (
         cd ${TMPBUILDDIR}/${MODFILE}-${VERS}
         if [ "x$TEST" = "xyes" ]; then
-          $PERL Makefile.PL
+          if [ "x$MODFILE" = "xNet-DNS" ]; then
+            # Net-DNS asks about wanting live tests, which we don't.
+            yes n | $PERL Makefile.PL
+          else
+            $PERL Makefile.PL
+          fi
           #[ "x$CCISGCC" = "xyes" ] && $PERL -pi.bak -e 's/-KPIC|-xO3|-xdepend//g' Makefile
           $MAKE && $MAKE test && $MAKE install
         else
