@@ -2,7 +2,7 @@
 #   MailScanner - SMTP E-Mail Virus Scanner
 #   Copyright (C) 2002  Julian Field
 #
-#   $Id: ConfigDefs.pl 4414 2008-04-22 09:06:02Z sysjkf $
+#   $Id: ConfigDefs.pl 4484 2008-06-17 19:45:00Z sysjkf $
 #
 #   This program is free software; you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -129,6 +129,7 @@ inlinetextsig			= inlinetextsignature
 inlinetextwarning		= inlinetextwarning
 inqueuedir			= incomingqueuedir
 ipverheader			= ipprotocolversionheader
+isareply			= dontsignhtmlifheadersexist
 keepspamarchiveclean		= keepspamandmcparchiveclean
 lastafterbatch			= alwayslookeduplastafterbatch
 lastlookup			= alwayslookeduplast
@@ -420,7 +421,6 @@ LogPermittedFilenames	0	no	0	yes	1
 LogPermittedFiletypes	0	no	0	yes	1
 LogPermittedFileMimetypes	0	no	0	yes	1
 LookForUU		0	no	0	yes	1
-MissingMailArchiveIs	1	file	0	directory	1
 MultipleHeaders		append	append	append	replace	replace	add	add
 NameModifySubject	start	no	0	yes	1	start	start	end	end
 NoticeFullHeaders	1	no	0	yes	1
@@ -526,6 +526,7 @@ EnvToHeader			X-MailScanner-Envelope-To:
 HighMCPSubjectText		{MCP?}
 HighSpamSubjectText		{Spam?}
 Hostname			the MailScanner
+IDHeader			X-MailScanner-ID:
 InfoHeader			
 InfoValue			Please contact the ISP for more information
 IPVerHeader			
@@ -613,6 +614,7 @@ ClamdLockFile
 ClamdSocket 127.0.0.1
 FilenameRules		
 FiletypeRules		
+isareply
 mshmacnull			spam
 NoisyViruses			Joke/ OF97/ WM97/ W97M/ eicar
 NoSenderPrecedence		list bulk
