@@ -2,7 +2,7 @@
 #   MailScanner - SMTP E-Mail Virus Scanner
 #   Copyright (C) 2002  Julian Field
 #
-#   $Id: Postfix.pm,v 1.1.2.45 2004/05/31 13:16:33 jkf Exp $
+#   $Id: Postfix.pm 4372 2008-04-02 15:45:26Z sysjkf $
 #
 #   This program is free software; you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -37,11 +37,12 @@ use strict 'refs';
 no  strict 'subs'; # Allow bare words for parameter %'s
 
 use DirHandle;
+use Encode;
 
 use vars qw($VERSION);
 
 ### The package version, both in 1.23 style *and* usable by MakeMaker:
-$VERSION = substr q$Revision: 1.1.2.45 $, 10;
+$VERSION = substr q$Revision: 4372 $, 10;
 
 # Command-line options you need to give to sendmail to sensibly process a
 # message that is piped to it. Still need to add "-f" for specifying the
@@ -71,7 +72,7 @@ sub initialise {
   MailScanner::Config::Default('sendmail', '/usr/sbin/sendmail');
   MailScanner::Config::Default('sendmail2',
                                MailScanner::Config::Value('sendmail'));
-  $MailScanner::SMDiskStore::HashDirDepth = 0;
+  $MailScanner::SMDiskStore::HashDirDepth = -1;
   $UnsortedBatchesLeft = 0; # Disable queue-clearing mode
 }
 
@@ -237,6 +238,7 @@ sub new {
     #$id =~ /^(.)(.)(.+)$/;
     #($dir1, $dir2, $file) = ($1,$2,$3);
     #return "$dir1/$dir2/$file";
+    $id =~ s/\.[^.]+$//;
     return "$id";
   }
 
@@ -261,6 +263,8 @@ sub new {
     } elsif ($MailScanner::SMDiskStore::HashDirDepth == 1) {
       $file =~ /^(.)/;
       return ($dir,$1,$file);
+    } elsif ($MailScanner::SMDiskStore::HashDirDepth == 0) {
+      return ($dir,$file);
     } else {
       MailScanner::Log::WarnLog("Postfix dir depth has not been set!");
     }
@@ -282,16 +286,29 @@ sub new {
     my($ip, $TOline);
     my($Line, $Flags);
     my($MsgContSize, $DataOffset, $NumRecips);
-    my($TOFound, $FROMFound, $IPFound);
-    my($rectype, $recdata);
+    # ORIGFound stuff courtesy of Juan Pablo Abuyeres 23/7/2006
+    # Should improve handling of virtual domains.
+    my($ORIGFound, $TOFound, $FROMFound, $IPFound, $TIMEFound);
+    my($ErrorFound, $ERecordFound, $rectype, $recdata, $mtime);
+    my $InSubject = 0; # Are we adding continuation subject lines?
+    my $pRecordsFound = 0; # p record spin-through of body? //Glenn
+    my $OriginalPos = 0; # p record jumpoff point //Glenn
+    my $MaxpRecPos = 0; # Max position where a p record might occur.
+    my @npos = (); # save all p record positions, although we only
+                   # look at/use the first and last.
 
     #print STDERR "In ReadQf\n";
     #$message->{store}->print();
     $message->{nobody} = 0; # If there is no message body we just get X at end of headers
+    $ERecordFound = 0;
 
     # Just in case we get a message with no headers at all
     @{$message->{headers}} = ();
 
+    # seek to end of file and save position, to make sure p records don't
+    # try go past this point. 17 is the record size for a p record.
+    seek $RQf, -17, 2;
+    $MaxpRecPos = tell $RQf;
 
     # Seek to the start of the file in case anyone read the file
     # between me opening it and locking it.
@@ -303,25 +320,46 @@ sub new {
     #print "1st $rectype is \"$recdata\"\n";
     MailScanner::Log::WarnLog("Syntax error in Postfix queue file, didn't " .
                               "start with a C record") unless $rectype eq 'C';
-    $recdata =~ /^([0-9 ]{15}) ([0-9 ]{15}) ([0-9 ]{15})( ([0-9 ]{15}))?$/;
-    ($MsgContSize, $DataOffset, $NumRecips) = ($1+0, $2+0, $3+0);
+    #$recdata =~ /^([0-9 ]{15}) ([0-9 ]{15}) ([0-9 ]{15})( ([0-9 ]{15}))?$/;
+    #($MsgContSize, $DataOffset, $NumRecips) = ($1+0, $2+0, $3+0);
+
+    my @numbers = split " ", $recdata;
+    ($MsgContSize, $DataOffset, $NumRecips) =
+      ($numbers[0]+0, $numbers[1]+0, $numbers[2]+0);
 
     # If $5 is set then we have a new data structure in the file
     $MailScanner::Postfix::DataStructure = 0;
-    if ($5 ne "") {
+    #if ($5 ne "") {
+    #  $MailScanner::Postfix::DataStructure = 1;
+    #  $message->{PostfixQmgrOpts} = $5+0;
+    #}
+
+    if (defined $numbers[3]) {
       $MailScanner::Postfix::DataStructure = 1;
-      $message->{PostfixQmgrOpts} = $5+0;
+      $message->{PostfixQmgrOpts} = $numbers[3]+0;
     }
 
-    $MsgContSize =~ s/^\s*//;
-    $DataOffset  =~ s/^\s*//;
-    $NumRecips   =~ s/^\s*//;
+    #$MsgContSize =~ s/^\s*//;
+    #$DataOffset  =~ s/^\s*//;
+    #$NumRecips   =~ s/^\s*//;
     #print STDERR "MsgContSize=$MsgContSize DataOffset=$DataOffset NumRecips=$NumRecips\n";
     push @{$message->{metadata}}, "$rectype$recdata";
     #print STDERR "Content size = $MsgContSize\n";
     #print STDERR "Data offset  = $DataOffset\n";
     #print STDERR "Num Recips   = $NumRecips\n";
 
+    # If the data offset is 0 then Postfix definitely hasn't finished
+    # writing the message.
+    unless ($DataOffset+0 > 10) { # 10 == arbitrary small number
+      # JKF 5/12/2005 This could fail with an unblessed reference error
+      # JKF 5/12/2005 so do it by hand.
+      # JKF 5/12/2005 $message->DropFromBatch();
+      $message->{deleted} = 1;
+      $message->{gonefromdisk} = 1; # Don't try to delete the original
+      $message->{store}->Unlock(); # Unlock it so other processes can pick it up
+      return 0;
+    }
+
     # Read records until we hit the start of the message M record
     #print STDERR "Reading pre data\n";
     while(($rectype, $recdata) = ReadRecord($RQf)) {
@@ -344,6 +382,7 @@ sub new {
         push @{$message->{to}}, lc($recdata);
         push @{$message->{metadata}}, "$rectype$recdata";
         $TOFound = 1;
+        $ORIGFound = 1;
         #print STDERR "Pre R Recip $recdata\n";
       #JKF 20040322 } elsif ($rectype eq 'O') {
       } elsif ($rectype eq 'R') {
@@ -354,12 +393,63 @@ sub new {
         $recdata =~ s/^\<//;
         $recdata =~ s/\<$//;
         $recdata = lc($recdata);
-        push @{$message->{to}}, $recdata;
+        push @{$message->{to}}, lc($recdata) unless $ORIGFound;
         push @{$message->{metadata}}, "$rectype$recdata";
         #JKF 20040322 $message->{originalrecips}{"$recdata"} = 1;
-        $message->{postfixrecips}{"$recdata"} = 1;
+        $message->{postfixrecips}{lc("$recdata")} = 1;
         $TOFound = 1;
         #print STDERR "Pre O Recip $recdata\n";
+      } elsif ($rectype eq 'p') {
+	# //Glenn 2007-01-16
+	# Handle p records (GOTO like in-place edit thing) by reading
+	# the pointed to data into the main message object, and
+	# silently discarding the actual p record.
+	# This p record should only point to added recipients, or
+	# moved records of the same type already handled in this
+	# segment, so lets just store the jumpoff point and loop.
+	# When we hit the next p record it should be the "jump back to
+        # original pos" one, or another forward p record... so we'll
+	# check that and act accordingly..
+        $pRecordsFound = 1; # If we find a p record we need remember this to handle body ...
+	next if ($recdata+0 == 0); # Ignore zero (placeholder) jumps
+	if ($recdata+0 > $MaxpRecPos) {
+	    MailScanner::Log::WarnLog("p record handling: Attempt to jump beyond end of file, aborting file.");
+	    $ErrorFound = 1;
+	    last;
+	}
+	push @npos, $recdata+0; # save jumpto position, for loop detection.
+	if ($OriginalPos == 0) {
+	   # Jump to after E record and commence
+	   $OriginalPos = tell $RQf;
+	   seek $RQf, $recdata+0, 0; # jump. we should check this works.
+	   next;
+	} else {
+	  # We're at the return point, or moving even furtehr away...
+	  if ($recdata+0 < $OriginalPos) {
+	     # Error exit, carp or whatever. Probably should do the
+	     # nice exit Jules does above. For now, warn and use
+	     # OriginalPos
+	     MailScanner::Log::WarnLog("p record handling: $recdata < $OriginalPos, using $OriginalPos instead!");
+	     seek $RQf, $OriginalPos, 0; # jump back up. we should chk this.
+	     $OriginalPos = 0; # We're home again... Reset to prepare for
+			       # any new warps
+	     @npos = ();
+	  } else {
+	    seek $RQf, $recdata+0, 0; # jump back or forward. we should chk this works.
+	    if ($recdata+0 < $npos[$#npos-1]) {
+	       if ($recdata+0 >= $npos[0]) {
+		  MailScanner::Log::WarnLog("p record handling: Loop condition found, aborting file.");
+		  $ErrorFound = 1;
+		  $OriginalPos = 0; # Reset to not fool next segment loop
+	          @npos = ();
+		  last;
+	       }
+	       $OriginalPos = 0; # We're home again... Reset to prepare for
+		                 # any new warps
+	       @npos = ();
+	    }
+	  }
+	}
       } else {
         # Some other record type. Just store it and move on.
         push @{$message->{metadata}}, "$rectype$recdata";
@@ -378,11 +468,77 @@ sub new {
         last;
       }
       last if $rectype eq 'N' && $recdata eq "";
+      if ($rectype eq 'p') {
+	 # //Glenn 2007-01-16
+	 # Handle p records (GOTO like in-place edit thing) by reading
+	 # the pointed to data into the main message object, and
+	 # silently discarding the actual p record.
+	 # This p record should only point to added headers, or
+	 # moved records of the same type already handled in this
+	 # segment, so lets just store the jumpoff point and loop.
+	 # When we hit the next p record it should be the "jump back to
+         # original pos" one, or another forward p record... so we'll
+	 # check that and act accordingly..
+         $pRecordsFound = 1; # If we find a p record we need remember this to handle body ...
+	 next if ($recdata+0 == 0); # Ignore zero (placeholder) jumps
+	 if ($recdata+0 > $MaxpRecPos) {
+	    MailScanner::Log::WarnLog("p record handling: Attempt to jump beyond end of file, aborting file.");
+	    $ErrorFound = 1;
+	    last;
+	 }
+	 push @npos, $recdata+0; # save jumpto position, for loop detection.
+	 if ($OriginalPos == 0) {
+	    # Jump to after E record and commence
+	    $OriginalPos = tell $RQf;
+	    seek $RQf, $recdata+0, 0; # jump. we should check this works.
+#	    next;
+	 } else {
+	   # We're at the return point, or moving even furtehr away...
+	   if ($recdata+0 < $OriginalPos) {
+	      # Error exit, carp or whatever. Probably should do the
+	      # nice exit Jules does above. For now, warn and use
+	      # OriginalPos
+	      MailScanner::Log::WarnLog("p record handling: $recdata < $OriginalPos, using $OriginalPos instead!");
+	      seek $RQf, $OriginalPos, 0; # jump back up. we should chk this.
+	      $OriginalPos = 0; # We're home again... Reset to prepare for
+			        # any new warps
+	      @npos = ();
+	   } else {
+	     seek $RQf, $recdata+0, 0; # jump back or forward. we should chk this works.
+	     if ($recdata+0 < $npos[$#npos-1]) {
+	        if ($recdata+0 >= $npos[0]) {
+        	   MailScanner::Log::WarnLog("p record handling: Loop condition found, aborting file.");
+		   $ErrorFound = 1;
+	           $OriginalPos = 0; # Reset to not fool next segment loop...
+	           @npos = ();
+		   last;
+	        }
+	        $OriginalPos = 0; # We're home again... Reset to prepare for
+		                  # any new warps
+	        @npos = ();
+	     }
+	   }
+	 }
+	 next; # done, don't add a spurious "converted p to N" record.
+      }
+      next if ($rectype eq 'w'); # Skip deleted (w) records ... Else they will be transformed to Normal (N) records.
       push @{$message->{headers}}, $recdata; # Headers have no leading N
-      if ($recdata =~ /^Subject:\s+(\S.*)$/i) {
+      if ($recdata =~ /^Subject:\s*(\S.*)?$/i) {
         $message->{subject} = $1;
+        $InSubject = 1;
         next;
       }
+      if ($InSubject) {
+        if ($recdata =~ /^\s/) {
+          # We are in a continuation line, so remove the leading whitespace
+          $recdata =~ s/^\s//;
+          $message->{subject} .= $recdata;
+          next;
+        } else {
+          # Line did not start with continuation character so we're not in Subj
+          $InSubject = 0;
+        }
+      }
       if (!$IPFound && $recdata =~ /^Received: .+\[(\d+\.\d+\.\d+\.\d+)\]/i) {
         $message->{clientip} = $1;
         $IPFound = 1;
@@ -400,7 +556,89 @@ sub new {
 
     # We are now at the end of the headers. Jump straight to the metadata
     # after the message.
-    seek $RQf, $MsgContSize+$DataOffset, 0;
+#    seek $RQf, $MsgContSize+$DataOffset, 0;
+
+# Inelegant, but working. Instead of an efficient seek, we spinn through to
+# after X record. Unless we don't have a body to spin through. Also skip
+# the spin if we don't have any p records already (to not punish the normal
+# case).
+    if (!$message->{nobody} && $pRecordsFound) {
+      while(($rectype, $recdata) = ReadRecord($RQf)) {
+        #print STDERR "Metadata type $rectype data \"$recdata\"\n";
+        if (!defined($rectype) or $rectype eq 'X') {
+	  $ErrorFound = 1 if(!defined($rectype));
+          last;
+        }
+        if ($rectype eq 'p') {
+	  # //Glenn 2007-01-16
+	  # Handle p records (GOTO like in-place edit thing) by reading
+	  # the pointed to data into the main message object, and
+	  # silently discarding the actual p record.
+	  # This p record should only point to a new body record, or
+	  # moved records of the same type already handled in this
+	  # segment, so lets just store the jumpoff point and loop.
+	  # When we hit the next p record it should be the "jump back to
+          # original pos" one, or another forward p record... so we'll
+	  # check that and act accordingly..
+	  next if ($recdata+0 == 0); # Ignore zero (placeholder) jumps
+	  if ($recdata+0 > $MaxpRecPos) {
+	      MailScanner::Log::WarnLog("p record handling: Attempt to jump beyond end of file, aborting file.");
+	      $ErrorFound = 1;
+	      last;
+	  }
+	  push @npos, $recdata+0; # save jumpto position, for loop detection.
+	  if ($OriginalPos == 0) {
+	     # Jump to after E record and commence
+	     $OriginalPos = tell $RQf;
+	     seek $RQf, $recdata+0, 0; # jump. we should check this works.
+	     next;
+	  } else {
+	    # We're at the return point, or moving even furtehr away...
+	    if ($recdata+0 < $OriginalPos) {
+	       # Error exit, carp or whatever. Probably should do the
+	       # nice exit Jules does above. For now, warn and use
+	       # OriginalPos
+	       MailScanner::Log::WarnLog("p record handling: $recdata < $OriginalPos, using $OriginalPos instead!");
+	       seek $RQf, $OriginalPos, 0; # jump back up. we should chk this.
+	       $OriginalPos = 0; # We're home again... Reset to prepare for
+			         # any mew warps
+	       @npos = ();
+	    } else {
+	      seek $RQf, $recdata+0, 0; # jump back or forward. we should chk this works.
+	      if ($recdata+0 < $npos[$#npos-1]) {
+	         if ($recdata+0 >= $npos[0]) {
+		    MailScanner::Log::WarnLog("p record handling: Loop condition found, aborting file.");
+		    $ErrorFound = 1;
+		    $OriginalPos = 0; # Reset to not fool next segment loop...
+		    last;
+	         }
+	         $OriginalPos = 0; # We're home again... Reset to prepare for
+		                   # any mew warps
+	         @npos = ();
+	      }
+	    }
+	  }
+        }
+      }
+      #print STDERR "\n\nErrorFound=$ErrorFound and rectype=$rectype\n\n";
+      # Found errors above, means the file isn't complete... remove it from the batch and return immediately.
+      if ($ErrorFound or $rectype ne 'X') {
+        #MailScanner::Log::WarnLog("No end-of-message record found in %s, " .
+        #                          "retrying", $message->{id});
+        # JKF 5/12/2005 This could fail with an unblessed reference error
+        # JKF 5/12/2005 so do it by hand.
+        # JKF 5/12/2005 $message->DropFromBatch();
+        $message->{deleted} = 1;
+        $message->{gonefromdisk} = 1; # Don't try to delete the original
+        $message->{store}->Unlock(); # Unlock it so other processes can pick it up
+
+        return 0;
+      }
+    }
+    # "safety" seek, in case things go badly above. We also need to return "before" the X record, so that it is copied over below.
+    #my $CurrentPos = tell $RQf;
+    #print STDERR "MsgContSize+DataOffset = ",$MsgContSize+$DataOffset,"\nCuurentPos = ",$CurrentPos+0,"\n";
+    seek $RQf, $MsgContSize+$DataOffset, 0; # if ($MsgContSize+$DataOffset ne $CurrentPos+0);
 
     # We are now in the metadata after the message.
     #print STDERR "Reading post data\n";
@@ -408,8 +646,17 @@ sub new {
       #print STDERR "Metadata type $rectype data \"$recdata\"\n";
       if ($rectype eq 'E') {
         push @{$message->{metadata}}, "$rectype$recdata";
+        $ERecordFound = 1;
         last;
       }
+      # JKF 20050621 Must only ever find 1 timestamp or the message is corrupt
+      if ($rectype eq 'T') {
+        if ($TIMEFound) {
+          $ErrorFound = 1;
+          last;
+        }
+        $TIMEFound = 1;
+      }
       #JKF 20040322 if ($rectype eq 'R') {
       if ($rectype eq 'O') {
         # Recipient address
@@ -418,6 +665,7 @@ sub new {
         push @{$message->{to}}, lc($recdata);
         push @{$message->{metadata}}, "$rectype$recdata";
         $TOFound = 1;
+        $ORIGFound = 1;
         #print STDERR "Post R Recip $recdata\n";
       #JKF 20040322 } elsif ($rectype eq 'O') {
       } elsif ($rectype eq 'R') {
@@ -427,18 +675,77 @@ sub new {
         $recdata =~ s/^\<//;
         $recdata =~ s/\<$//;
         $recdata = lc($recdata);
-        push @{$message->{to}}, $recdata;
+        #push @{$message->{to}}, $recdata;
+        push @{$message->{to}}, lc($recdata) unless $ORIGFound;
         push @{$message->{metadata}}, "$rectype$recdata";
         #JKF 20040322 $message->{originalrecips}{"$recdata"} = 1;
         $message->{postfixrecips}{"$recdata"} = 1;
         $TOFound = 1;
         #print STDERR "Post O Recip $recdata\n";
+      } elsif ($rectype eq 'p') {
+	# //Glenn 2007-01-16
+	# Handle p records (GOTO like in-place edit thing) by reading
+	# the pointed to data into the main message object, and
+	# silently discarding the actual p record.
+	# This p record should only point to added recipients, or
+	# moved records of the same type already handled in this
+	# segment, so lets just store the jumpoff point and loop.
+	# When we hit the next p record it should be the "jump back to
+        # original pos" one, or another forward p record... so we'll
+	# check that and act accordingly..
+	# I'm not sure this segment can have p records, but better
+	# safe than sorry.
+	next if ($recdata+0 == 0); # Ignore zero (placeholder) jumps
+	if ($recdata+0 > $MaxpRecPos) {
+	    MailScanner::Log::WarnLog("p record handling: Attempt to jump beyond end of file, aborting file.");
+	    $ErrorFound = 1;
+	    last;
+	}
+	push @npos, $recdata+0; # save jumpto position, for loop detection.
+	if ($OriginalPos == 0) {
+	   # Jump to after E record and commence
+	   $OriginalPos = tell $RQf;
+	   seek $RQf, $recdata+0, 0; # jump. we should check this works.
+	   next;
+	} else {
+	  # We're at the return point, or moving even furtehr away...
+	  if ($recdata+0 < $OriginalPos) {
+	     # Error exit, carp or whatever. Probably should do the
+	     # nice exit Jules does above. For now, warn and use
+	     # OriginalPos
+	     MailScanner::Log::WarnLog("p record handling: $recdata < $OriginalPos, using $OriginalPos instead!");
+	     seek $RQf, $OriginalPos, 0; # jump back up. we should chk this.
+	     $OriginalPos = 0; # We're home again... Reset to prepare for
+			       # any mew warps
+	     @npos = ();
+	  } else {
+	    seek $RQf, $recdata+0, 0; # jump back or forward. we should chk this works.
+	    if ($recdata+0 < $npos[$#npos-1]) {
+	       if ($recdata+0 >= $npos[0]) {
+		  MailScanner::Log::WarnLog("p record handling: Loop condition found, aborting file.");
+		  $ErrorFound = 1;
+		  # no need to reset $OriginalPos since this is the last segment.
+		  last;
+	       }
+	       $OriginalPos = 0; # We're home again... Reset to prepare for
+		                 # any mew warps
+	       @npos = ();
+	    }
+	  }
+	}
       } else {
         # Some other record type. Just store it and move on.
         push @{$message->{metadata}}, "$rectype$recdata";
       }
     }
       
+    # Remove all the duplicates from ->{to}
+    my %uniqueto;
+    foreach (@{$message->{to}}) {
+      $uniqueto{$_} = 1;
+    }
+    @{$message->{to}} = keys %uniqueto;
+
     # We now have all the pre-message records followed by the M record
     # followed by the post-message records including the X record and the
     # terminating E record. We can add recipient R records just before
@@ -457,7 +764,51 @@ sub new {
     # we have something there
     $message->{clientip} = '0.0.0.0' unless $IPFound;
 
-    return 1 if $FROMFound && $TOFound; # && $IPFound;
+    # Decode ISO subject lines into UTF8
+    # Needed for UTF8 support in MailWatch 2.0
+    eval {
+     $message->{utf8subject} = Encode::decode('MIME-Header',
+                                              $message->{subject});
+    };
+    if ($@) {
+     # Eval failed - store a copy of the subject before MIME::WordDecoder
+     # is run, as this appears to destroy the characters of some subjects
+     $message->{utf8subject} = $message->{subject};
+    }
+
+    # Decode the ISO encoded Subject line
+    # Over-ride the default default character set handler so it does it
+    # much better than the MIME-tools default handling.
+    MIME::WordDecoder->default->handler('*' => \&MailScanner::Message::WordDecoderKeep7Bit);
+    my $TmpSubject = MIME::WordDecoder::unmime($message->{subject});
+    if ($TmpSubject ne $message->{subject}) {
+      # The unmime function dealt with an encoded subject, as it did
+      # something. Allow up to 10 trailing spaces so that SweepContent
+      # is more kind to us and doesn't go and replace the whole subject,
+      # thinking that it is malicious. Total replacement and hence
+      # destruction of unicode subjects is rather harsh when we are just
+      # talking about a few spaces.
+      $TmpSubject =~ s/ {1,10}$//;
+      $message->{subject} = $TmpSubject;
+    }
+    #old $message->{subject} = MIME::WordDecoder::unmime($message->{subject});
+
+    # If we never found the E (end of message file) record, then
+    # Postfix is definitely still writing the message.
+    unless ($ERecordFound) {
+      MailScanner::Log::WarnLog("No end-of-message record found in %s, " .
+                                "retrying", $message->{id});
+      # JKF 5/12/2005 This could fail with an unblessed reference error
+      # JKF 5/12/2005 so do it by hand.
+      # JKF 5/12/2005 $message->DropFromBatch();
+      $message->{deleted} = 1;
+      $message->{gonefromdisk} = 1; # Don't try to delete the original
+      $message->{store}->Unlock(); # Unlock it so other processes can pick it up
+
+      return 0;
+    }
+
+    return 1 if $FROMFound && $TOFound && !$ErrorFound; # && $IPFound;
     #MailScanner::Log::WarnLog("Batch: Found invalid queue file for " .
     #                          "message %s", $message->{id});
     return 0;
@@ -598,6 +949,35 @@ sub new {
     }
   }
 
+  sub UniqHeader {
+    my($this, $message, $key) = @_;
+
+    my($pos, $foundat);
+    $pos = 0;
+    $pos++ while ($message->{metadata}[$pos] !~ /^M/);
+    # Now points at the M record
+    $pos++;
+    # Now points at first N record
+    $foundat = -1;
+    while ($pos < @{$message->{metadata}}) {
+      if ($message->{metadata}[$pos] =~ /^N$key/i) {
+        if ($foundat == -1) { # Skip 1st occurrence
+          $foundat = $pos;
+          $pos++;
+          next;
+        }
+        # We have found the start of 1 occurrence of this header
+        splice @{$message->{metadata}}, $pos, 1;
+        # Delete continuation lines
+        while($message->{metadata}[$pos] =~ /^N\s/) {
+          splice @{$message->{metadata}}, $pos, 1;
+        }
+        next;
+      }
+      $pos++;
+    }
+  }
+
   sub ReplaceHeader {
     my($this, $message, $key, $newvalue) = @_;
 
@@ -680,7 +1060,7 @@ sub new {
       return;
     }
 
-    $message->{metadata}[$oldlocation] =~ s/^N$key\s+/N$key $newvalue$sep/i;
+    $message->{metadata}[$oldlocation] =~ s/^N$key\s*/N$key $newvalue$sep/i;
   }
 
   sub TextStartsHeader {
@@ -756,6 +1136,13 @@ sub new {
     my $this = shift;
     my($message, @recips) = @_;
 
+    # Remove all the duplicates @recips
+    my %uniqueto;
+    foreach (@recips) {
+      $uniqueto{$_} = 1;
+    }
+    @recips = keys %uniqueto;
+
     my $totallines = @{$message->{metadata}};
 
     foreach (@recips) {
@@ -779,7 +1166,11 @@ sub new {
     for ($linenum=0; $linenum<@{$message->{metadata}}; $linenum++) {
       # Looking for "recipient" lines
       # Should allow 'O' here as well
-      next unless $message->{metadata}[$linenum] =~ /^[RO]/;
+      # JKF 30/08/2006 next unless $message->{metadata}[$linenum] =~ /^[RO]/;
+      # Thanks to Holger Gebhard for this.
+      #BUGGY: next unless $message->{metadata}[$linenum] =~ /^[ARO].+@(?:\w|-|\.)+\.\w{2,})/;
+      #next unless $message->{metadata}[$linenum] =~ /^[ARO]/;
+      next unless $message->{metadata}[$linenum] =~ /^[ARO].+@(?:\w|-|\.)+\.\w{2,}/;
       # Have found the right line
       #print STDERR "Deleting recip " . $message->{metadata}[$linenum] . "\n";
       splice(@{$message->{metadata}}, $linenum, 1);
@@ -790,22 +1181,28 @@ sub new {
 
   # Send an I down the FIFO to the Postfix queue manager, so that it reads
   # its incoming queue.
+  # I am passed a hash of queues --> space-separated string of message ids
   sub KickMessage {
-    my(@ids) = @_;
-    my($idlist);
+    my($queue2ids, $sendmail2) = @_;
+    my($queue);
 
-    # Using the spool directory with the last element chopped off,
-    # find the public directory wth the qmgr FIFO in it. Send an I
-    # to that FIFO.
-    my $public = MailScanner::Config::Value('outqueuedir');
-    $public =~ s/[^\/]+$/public/;
-    my $fh = new FileHandle;
-    $fh->open(">$public/qmgr") or
-      MailScanner::Log::WarnLog("KickMessage failed as couldn't write to " .
-                                "%s, %s", "$public/qmgr", $!);
-    print $fh "I";
-    $fh->close;
+    # Do a kick for every queue that contains some message ids
+    foreach $queue (keys %$queue2ids) {
+      next unless $queue2ids->{$queue};
 
+      # Using the spool directory with the last element chopped off,
+      # find the public directory wth the qmgr FIFO in it. Send an I
+      # to that FIFO.
+      my $public = $queue;
+      $public =~ s/[^\/]+$/public/;
+      next unless $public; # Sanity checking!
+      my $fh = new FileHandle;
+      $fh->open(">$public/qmgr") or
+        MailScanner::Log::WarnLog("KickMessage failed as couldn't write to " .
+                                  "%s, %s", "$public/qmgr", $!);
+      print $fh "I";
+      $fh->close;
+    }
     return 0;
   }
 
@@ -824,13 +1221,16 @@ sub new {
     my($message) = @_;
 
     my($linenum, $result, $type, $data, $to, $preNlen);
+    my $TimestampFound = 0;
 
     #print STDERR "In PreDataString\n";
     # Output all the metadata records up until (& including) the M record.
     $linenum = 0;
+    $result  = '';
     foreach (@{$message->{metadata}}) {
       /^(.)(.*)$/;
       ($type, $data) = ($1, $2);
+      $TimestampFound++ if $type eq 'T'; # Must only ever have 1 timestamp
       #print STDERR "PreData1 Type $type Data $data\n";
       last if $type eq 'M';
       $result .= Record2String($type, $data);
@@ -864,10 +1264,11 @@ sub new {
     # Store the length of th estring so far as we need to return it
     $preNlen = length($result);
 
+    my $totallines = scalar(@{$message->{metadata}});
     # Add the headers
-    foreach(; $linenum<@{$message->{metadata}}; $linenum++) {
-      $_ = $message->{metadata}[$linenum];
-      /^(.)(.*)$/;
+    for ($linenum=$linenum; $linenum<$totallines; $linenum++) {
+      #$_ = $message->{metadata}[$linenum];
+      $message->{metadata}[$linenum] =~ /^(.)(.*)$/;
       ($type, $data) = ($1, $2);
       #print STDERR "PreData2 Type $type Data $data\n";
       last if $type eq 'X';
@@ -881,7 +1282,7 @@ sub new {
 
     #print STDERR "Result of PreDataString is $result\n";
     # Return the string and the length of the data before any N records
-    return ($result, $preNlen);
+    return ($result, $preNlen, $TimestampFound);
   }
 
   sub PostDataString {
@@ -890,6 +1291,7 @@ sub new {
 
     my($result, $type, $data);
     my($record, $recordnum);
+    my $TimestampFound = 0;
     $result = Record2String('X', "");
 
     $recordnum = @{$message->{metadata}} - 1;
@@ -899,10 +1301,11 @@ sub new {
       $record =~ /^(.)(.*)$/;
       ($type, $data) = ($1, $2);
       $result .= Record2String($type, $data);
+      $TimestampFound++ if $type eq 'T';
       #print STDERR "Post $type $data\n";
     }
 
-    return $result;
+    return($result, $TimestampFound);
   }
 
 
@@ -925,6 +1328,22 @@ sub new {
   }
 
 
+  # Append, add or replace a given header with a given value.
+  sub AddMultipleHeaderName {
+    my $this = shift;
+    my($message, $headername, $headervalue, $separator) = @_;
+
+    my($multiple) = MailScanner::Config::Value('multipleheaders', $message);
+    $this->AppendHeader ($message, $headername, $headervalue, $separator)
+      if $multiple eq 'append';
+
+    $this->AddHeader    ($message, $headername, $headervalue)
+      if $multiple eq 'add';
+
+    $this->ReplaceHeader($message, $headername, $headervalue)
+      if $multiple eq 'replace';
+  }
+
   # Append, add or replace a given header with a given value.
   sub AddMultipleHeader {
     my $this = shift;
@@ -1011,7 +1430,7 @@ sub FindHashDirDepth {
     while(defined($dir1name = $here->read())) {
       #MailScanner::Log::WarnLog("JKF: Reading %s from dir .", $dir1name);
       next if $dir1name eq '.' || $dir1name eq '..';
-      $filecount++ if -f $dir1name;
+      $filecount++ if -f $dir1name && $dir1name =~ /$mta->{HDFileRegexp}/;
       next unless -d $dir1name;
       $dir1->open($dir1name)
         or MailScanner::Log::DieLog("Cannot open dir %s when finding depth",
@@ -1041,7 +1460,8 @@ sub FindHashDirDepth {
 
     # Didn't find anything at all, so sleep waiting for a file or a dir
     # to appear in the queue.
-    MailScanner::Log::WarnLog("Messages found but no hashed queue directories. Please enable hashed queues for incoming and deferred with a depth of 1 or 2. See the Postfix documentation for hash_queue_names and hash_queue_depth")
+    # Can now be 0, 1 or 2: MailScanner::Log::WarnLog("Messages found but no hashed queue directories. Please enable hashed queues for incoming and deferred with a depth of 1 or 2. See the Postfix documentation for hash_queue_names and hash_queue_depth")
+    return 0
       if $filecount>0;
     sleep($delay);
   }
@@ -1052,7 +1472,7 @@ sub FindHashDirDepth {
   # the passed-in batch object.
   sub CreateBatch {
     my $this = shift;
-    my($batch) = @_;
+    my($batch, $onlyid) = @_;
 
     my($queuedirname, $queuedir, $queue1dir, $queue2dir, $MsgsInQueue);
     my($DirtyMsgs, $DirtyBytes, $CleanMsgs, $CleanBytes);
@@ -1061,7 +1481,7 @@ sub FindHashDirDepth {
     my(%ModDate, $mta, $file, $file1, $file2, $tmpdate);
     my(@SortedFiles, $id, $newmessage, @queuedirnames);
     my($batchempty, $h1, $h2, $delay, $CriticalQueueSize);
-    my($nlinks, $headerfileumask, $invalidfiles);
+    my($nlinks, $headerfileumask, $invalidfiles, $mtime);
 
     $queuedir  = new DirHandle;
     $queue1dir = new DirHandle;
@@ -1105,7 +1525,11 @@ sub FindHashDirDepth {
       $invalidfiles = "";
 
       # Loop through each of the inq directories
-      foreach $queuedirname (@queuedirnames) {
+      # Patch to combat starving in emergency queue mode
+      # foreach $queuedirname (@queuedirnames) {
+      my @aux_queuedirnames=@queuedirnames;
+      while( defined($queuedirname=splice(@aux_queuedirnames,
+            ($UnsortedBatchesLeft<=0 ? 0 :int(rand(@aux_queuedirnames))),1))) {
         #print STDERR "Scanning dir $queuedirname\n";
         unless (chdir $queuedirname) {
           MailScanner::Log::WarnLog("Cannot cd to dir %s to read messages, %s",
@@ -1117,7 +1541,7 @@ sub FindHashDirDepth {
         # If we haven't found the hash directory depth yet, then work it out
         #MailScanner::Log::WarnLog("JKF: About to work out hash dir depth");
         $MailScanner::SMDiskStore::HashDirDepth = FindHashDirDepth($mta)
-          unless $MailScanner::SMDiskStore::HashDirDepth;
+          unless $MailScanner::SMDiskStore::HashDirDepth >= 0;
         #MailScanner::Log::WarnLog("JKF: Hash dir depth is %d",
         #  $MailScanner::SMDiskStore::HashDirDepth);
 
@@ -1130,6 +1554,22 @@ sub FindHashDirDepth {
         # files in the the child directories.
         while(defined($file = $queuedir->read())) {
           next if $file eq '.' || $file eq '..';
+          if ($MailScanner::SMDiskStore::HashDirDepth==0) {
+            next unless $file =~ /$mta->{HDFileRegexp}/;
+            push @SortedFiles, "$queuedirname/$file";
+            if ($UnsortedBatchesLeft<=0) {
+              # Running normally
+              ($nlinks, $tmpdate) = (stat($file))[3,9]; # 9 = mtime
+              next if -z _;
+              next unless -f _;
+              next unless -R _;
+              next if $nlinks>1; # Catch files being moved into "deferred"
+              $ModDate{"$queuedirname/$file"} = $tmpdate;
+            }
+            $MsgsInQueue++;
+            #print STDERR "Stored depth 0 message file $file\n";
+            next;
+          }
           next unless -d $file;
           $queue1dir->open($file) or next;
           while(defined($file1 = $queue1dir->read())) {
@@ -1202,11 +1642,13 @@ sub FindHashDirDepth {
 
         # In accelerated queue-clearing mode, so we don't know anything yet
         if ($UnsortedBatchesLeft>0) {
-          $nlinks = (stat $file)[3];
+          ($nlinks, $mtime) = (stat $file)[3,9];
           next if -z _; # Skip 0-length queue files
           next unless -f _;
           next unless -R _;
           next if $nlinks>1; # Files being moved into "deferred"
+        } else {
+          $mtime = $ModDate{$file};
         }
 
         # Yes I know this is a hack but it will help isolate the problem
@@ -1222,24 +1664,38 @@ sub FindHashDirDepth {
         ($queuedirname, $h1, $file) = ($1,$2,$3)
           if $MailScanner::SMDiskStore::HashDirDepth == 1 &&
              $file =~ /^(.*)\/(.)\/([^\/]+)$/;
+        ($queuedirname, $file) = ($1,$2)
+          if $MailScanner::SMDiskStore::HashDirDepth == 0 &&
+             $file =~ /^(.*)\/([^\/]+)$/;
         next unless $file =~ /$mta->{HDFileRegexp}/;
-        $id = $1;
+        # Put the real message id in $idorig and the unique name in $id
+        # JKF Add a dot followed by a random number to try to get a unique
+        # JKF filename, as Postfix re-uses filenames too often.
+        my $idtemp = $1;
+        # If they want a particular message id, ignore it if it doesn't match
+        next if $onlyid ne "" && $idtemp ne $onlyid;
+
+        my $id = $idtemp . sprintf(".%05X", int(rand 1000000)+1);
+        my $idorig = $idtemp;
 
         #print STDERR "Adding $id to batch\n";
 
         # Lock and read the qf file. Skip this message if the lock fails.
         $newmessage = MailScanner::Message->new($id, $queuedirname);
         if ($newmessage eq 'INVALID') {
-          $invalidfiles .= "$id ";
+          #$invalidfiles .= "$id ";
+          $invalidfiles .= "$idorig ";
           next;
         }
         next unless $newmessage;
 
         $batch->{messages}{"$id"} = $newmessage;
+        $newmessage->{mtime} = $mtime;
         #print STDERR "Added $id to batch\n";
         $batchempty = 0;
 
-        if (MailScanner::Config::Value("virusscan", $newmessage)) {
+        if (MailScanner::Config::Value("virusscan", $newmessage) =~ /1/ ||
+            MailScanner::Config::Value("dangerscan", $newmessage) =~ /1/) {
           $newmessage->NeedsScanning(1);
           $DirtyMsgs++;
           $DirtyBytes += $newmessage->{size};
@@ -1268,7 +1724,7 @@ sub FindHashDirDepth {
     } while $batchempty; # Keep trying until we get something
 
     # Log the number of invalid messages found
-    MailScanner::Log::InfoLog("New Batch: Found invalid queue files: %s",
+    MailScanner::Log::NoticeLog("New Batch: Found invalid queue files: %s",
                               $invalidfiles)
       if $invalidfiles;
     # Log the size of the queue if it is more than 1 batch
