@@ -2,7 +2,7 @@
 #   MailScanner - SMTP E-Mail Virus Scanner
 #   Copyright (C) 2002  Julian Field
 #
-#   $Id: SweepOther.pm 4387 2008-04-07 14:47:48Z sysjkf $
+#   $Id: SweepOther.pm 4453 2008-05-21 20:04:02Z sysjkf $
 #
 #   This program is free software; you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
@@ -43,7 +43,7 @@ use POSIX qw(:signal_h setsid); # For Solaris 9 SIG bug workaround
 use vars qw($VERSION);
 
 ### The package version, both in 1.23 style *and* usable by MakeMaker:
-$VERSION = substr q$Revision: 4387 $, 10;
+$VERSION = substr q$Revision: 4453 $, 10;
 
 # Attributes are
 #
@@ -179,6 +179,9 @@ sub ScanBatch {
       $denypatterns  =~ s/\s+$//;
       @allowpatterns = split(" ", $allowpatterns);
       @denypatterns  = split(" ", $denypatterns);
+      # JKF Addenbrookes - Catch leading "*" on regexps and replace with ".*"
+      @allowpatterns = map { s/^\*/.\*/; $_; } @allowpatterns;
+      @denypatterns  = map { s/^\*/.\*/; $_; } @denypatterns;
       $allowexists   = @allowpatterns; # Don't use them if they are empty!
       $denyexists    = @denypatterns;
       $megaallow     = '(' . join(')|(',@allowpatterns) . ')';
@@ -264,6 +267,8 @@ sub ScanBatch {
           # Skip this rule if the regexp doesn't match
           # Check both filenames, the safe and the nasty. This is for
           # TNEF messages when the nasty filename is always winmail.dat
+          # JKF Addenbrookes - Catch leading "*" on regexps and replace with ".*"
+          $regexp =~ s/^\*/.\*/;
           next unless $attach =~ /$regexp/is || $safename =~ /$regexp/is;
           $MatchFound = 1;
 
@@ -519,6 +524,9 @@ sub CheckFileTypesRules {
     #$denypatterns  = quotemeta $denypatterns;  # quoting everything!
     @allowpatterns = split(" ", $allowpatterns);
     @denypatterns  = split(" ", $denypatterns);
+    # JKF Addenbrookes - Catch leading "*" on regexps and replace with ".*"
+    @allowpatterns = map { s/^\*/.\*/; $_; } @allowpatterns;
+    @denypatterns  = map { s/^\*/.\*/; $_; } @denypatterns;
     $allowexists   = @allowpatterns; # Don't use them if they are empty!
     $denyexists    = @denypatterns;
     $megaallow     = '(' . join(')|(',@allowpatterns) . ')';
@@ -579,6 +587,8 @@ sub CheckFileTypesRules {
 
         next if $regexp eq '' || $regexp eq '-'; # Skip this rule if there's no regexp at all
         #print STDERR "Type = $type, regexp = $regexp\n";
+        # JKF Addenbrookes - Catch leading "*" on regexps and replace with ".*"
+        $regexp =~ s/^\*/.\*/;
         next unless $type =~ /$regexp/is;
         #print STDERR "Filetype match: $allowdeny $regexp $logtext $usertext\n";
 
@@ -646,6 +656,9 @@ sub CheckFileTypesRules {
     #$denypatterns  = quotemeta $denypatterns;  # quoting everything!
     @allowpatterns = split(" ", $allowpatterns);
     @denypatterns  = split(" ", $denypatterns);
+    # JKF Addenbrookes - Catch leading "*" on regexps and replace with ".*"
+    @allowpatterns = map { s/^\*/.\*/; $_; } @allowpatterns;
+    @denypatterns  = map { s/^\*/.\*/; $_; } @denypatterns;
     $allowexists   = @allowpatterns; # Don't use them if they are empty!
     $denyexists    = @denypatterns;
     $megaallow     = '(' . join(')|(',@allowpatterns) . ')';
@@ -687,7 +700,7 @@ sub CheckFileTypesRules {
           $logtext = MailScanner::Config::LanguageValue($message,
                                                         'foundblockedfiletype');
           $usertext = $logtext;
-          MailScanner::Log::InfoLog("Filename Checks: %s (%s %s)",
+          MailScanner::Log::InfoLog("Filetype Checks: %s (%s %s)",
                                     $logtext, $id, $attach);
           $message->{namereports}{$safename} .= "$usertext ($safename)\n";
           $message->{nametypes}{$safename}   .= "f";
@@ -706,6 +719,8 @@ sub CheckFileTypesRules {
 
         next if $iregexp eq '' || $iregexp eq '-'; # Skip this rule if there's no regexp at all
         #print STDERR "iType = $type, iregexp = $iregexp\n";
+        # JKF Addenbrookes - Catch leading "*" on regexps and replace with ".*"
+        $iregexp =~ s/^\*/.\*/;
         next unless $type =~ /$iregexp/is;
         #print STDERR "iFiletype match: $allowdeny $regexp $iregexp $logtext $usertext ($type)\n";
 
